#!/usr/bin/env python

from homie.device_base import Device_Base
from homie.node.node_contact import Node_Contact
from homie.node.property.property_contact import Property_Contact
import logging

logger = logging.getLogger(__name__)


class Device_Contact(Device_Base):
    def __init__(
        self, device_id=None, name=None, homie_settings=None, mqtt_settings=None
    ):

        super().__init__(device_id, name, homie_settings, mqtt_settings)

        self.add_node(Node_Contact(self, id="contact"))

        self.start()

    def update_contact(self, state):
        self.get_node("contact").update_contact(state)
        logger.debug("Contact Update {}".format(state))

    def publish_homeassistant(self):
        hass_config = f'homeassistant/sensor/{self.device_id}/config'
        hass_payload = f'{{"name": "{self.name}","state_topic": "homie/{self.device_id}/contact/contact"}}'
        super().publish_homeassistant(hass_config,hass_payload)



from .property_float import Property_Float

tags = ["Lighting", "Dimmable"]


class Property_Dimmer(Property_Float):
    def __init__(
        self,
        node,
        id="dimmer",
        name="Dimmer",
        settable=True,
        retained=True,
        qos=1,
        unit="%",
        data_type=None,
        data_format="0:100",
        value=None,
        set_value=None,
        tags=tags,
        meta={},
    ):

        super().__init__(
            node,
            id,
            name,
            settable,
            retained,
            qos,
            unit,
            data_type,
            data_format,
            value,
            set_value,
            tags,
            meta,
        )

    def get_value_from_payload(self, payload):
        try:
            return round(float(payload),0)
        except:
            return None

    def get_payload_from_value(self, value):
        try:
            return round(value,2)
        except:
            return None


"""
from .property_integer import Property_Integer

tags = ['Lighting','Dimmable']

class Property_Dimmer(Property_Integer):

    def __init__(self,node, id='dimmer', name = 'Dimmer', settable = True, retained = True, qos=1, unit = '%', data_type= None, data_format = '0:100', value = None, set_value=None, tags=tags, meta={}):
        
        super().__init__(node,id,name,settable,retained,qos,unit,data_type,data_format,value,set_value,tags,meta)
"""




#!/usr/bin/env python

from homie.device_base import Device_Base
from homie.node.node_switch import Node_Switch
import logging

logger = logging.getLogger(__name__)


class Device_Switch(Device_Base):
    def __init__(
        self, device_id=None, name=None, homie_settings=None, mqtt_settings=None
    ):
        super().__init__(device_id, name, homie_settings, mqtt_settings)

        self.add_node(Node_Switch(self, id="switch", set_switch=self.set_switch))

        self.start()

    def update_switch(self, onoff):  # sends updates to clients
        self.get_node("switch").update_switch(onoff)
        logger.debug("Switch Update {}".format(onoff))

    def set_switch(self, onoff):  # received commands from clients
        logger.debug("Switch Set {}".format(onoff))


    def publish_homeassistant(self):
        hass_config = f'homeassistant/switch/{self.device_id}/config'
        hass_payload =  f'{{"name": "{self.name}","command_topic": "homie/{self.device_id}/switch/switch/set","state_topic": "homie/{self.device_id}/switch/switch","state_on" : "true","state_off" : "false","payload_on" : "true","payload_off" : "false"}}'

        super().publish_homeassistant(hass_config,hass_payload)




from .node_base import Node_Base

from homie.node.property.property_boolean import Property_Boolean


class Node_Boolean(Node_Base):
    def __init__(
        self,
        device,
        id="boolean",
        name="Boolean",
        type_="boolean",
        retain=True,
        qos=1,
        set_boolean=None,
    ):
        super().__init__(device, id, name, type_, retain, qos)

        assert set_boolean  # must provide a function to set the value of the switch

        self.add_property(Property_Boolean(self, set_value=set_boolean))

    def update_boolean(self, boolean):
        self.get_property("boolean").value = boolean




from .property_base import Property_Base


class Property_Boolean(Property_Base):
    def __init__(
        self,
        node,
        id="boolean",
        name="Boolean",
        settable=True,
        retained=True,
        qos=1,
        unit=None,
        data_type="boolean",
        data_format=None,
        value=None,
        set_value=None,
        tags=[],
        meta={},
    ):
        super().__init__(
            node,
            id,
            name,
            settable,
            retained,
            qos,
            unit,
            "boolean",
            data_format,
            value,
            set_value,
            tags,
            meta,
        )

    def validate_value(self, value):
        return isinstance(value, bool)

    def get_value_from_payload(self, payload):
        if payload == "true":
            return True
        elif payload == "false":
            return False
        else:
            return None

    def get_payload_from_value(self, value):  # convert value to a payload for homie
        return value and "true" or "false"




#!/usr/bin/env python

import sys
import time
import atexit
import signal
from datetime import datetime

import homie

from homie.support.helpers import validate_id
from homie.mqtt.homie_mqtt_client import connect_mqtt_client
from homie.mqtt.homie_mqtt_client import close_mqtt_clients
from homie.support.repeating_timer import Repeating_Timer

import logging

logger = logging.getLogger(__name__)


instance_count = 0  # used to track the number of device instances to allow for changing the default device id
devices = []

repeating_timer = None  # use common timer between all devices for updating state

DEVICE_STATES = [
    "init",
    "ready",
    "disconnected",
    "sleeping",
    "alert",
    "lost",
]

HOMIE_SETTINGS = {
    "version": "4.0.0",
    "topic": "homie",
    "fw_name": "homie4",
    "fw_version": homie.__version__,
    "update_interval": 60,
    "implementation": sys.platform,
}

EXTENSIONS = ["stats", "firmware", "meta"]  # extension we suport

EXTENSION_IDENTIFIERS = {
    "stats": "org.homie.legacy-stats:0.1.1:[4.x]",
    "firmware": "org.homie.legacy-firmware:0.1.1:[4.x]",
    "meta": "eu.epnw.meta:1.1.0:[3.0.1;4.x]",
}


class Device_Base(object):
    def __init__(
        self,
        device_id=None,
        name=None,
        homie_settings={},
        mqtt_settings={},
        extensions=["stats", "firmware", "meta"],
    ):
        global instance_count
        instance_count = instance_count + 1
        self.instance_number = instance_count

        self._mqtt_connected = False  # mqtt connection status, true = connected

        if device_id is None:
            device_id = self.generate_device_id()

        assert validate_id(device_id), "Invalid device id {}".format(device_id)
        self.device_id = device_id

        assert name
        self.name = name

        for extension in extensions:
            assert extension in EXTENSIONS
        self.extensions = extensions

        self._state = "init"

        self.homie_settings = self._homie_validate_settings(homie_settings)
        self.topic = "/".join((self.homie_settings["topic"], self.device_id))

        # may need a way to set these
        self.retained = True
        self.qos = 1

        self.nodes = {}

        self.start_time = None

        self.nodes_published = False

        self.mqtt_client = connect_mqtt_client(self, mqtt_settings)

        self.mqtt_subscription_handlers = {}

        global devices 
        devices.append(self)

        #atexit.register(self.close)

        #signal.signal(signal.SIGTERM, self.close)
        #signal.signal(signal.SIGINT, self.close)

    def close(self, *args):
        logger.info("Device Close {}".format(self.name))
        self.state = "disconnected"

    def generate_device_id(self):
        # logger.debug ('Device instances {}'.format(instance_count))
        # return "{:02x}".format(get_mac())+"{:04d}".format(instance_count)
        return "device{:04d}".format(self.instance_number)

    def start(self):  # called after the device has been built with nodes and properties
        logger.info("Device Start {}".format(self.name))
        self.start_time = time.time()

        if "stats" in self.extensions:
            global repeating_timer
            if repeating_timer == None:
                repeating_timer = Repeating_Timer(
                    self.homie_settings["update_interval"]
                )

            repeating_timer.add_callback(self.publish_uptime)

        if (
            self.mqtt_client.mqtt_connected
        ):  # run start up tasks if mqtt is ready, else wait for on_connect message from mqtt client
            self.mqtt_on_connection(True)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state, retain=True, qos=1):
        if state in DEVICE_STATES:
            self._state = state
            self.publish("/".join((self.topic, "$state")), self._state, retain, qos)
        else:
            logger.warning("Homie Invalid device state {}".format(state))

    def publish_attributes(self, retain=True, qos=1):
        self.publish(
            "/".join((self.topic, "$homie")),
            self.homie_settings["version"],
            retain,
            qos,
        )
        self.publish("/".join((self.topic, "$name")), self.name, retain, qos)
        self.publish(
            "/".join((self.topic, "$implementation")),
            self.homie_settings["implementation"],
            retain,
            qos,
        )

        self.publish_extensions(retain, qos)

        self.state = "ready"

    def publish_extensions(self, retain=True, qos=1):
        extensions = ",".join(self.extensions)

        if extensions is not "":
            self.publish("/".join((self.topic, "$extensions")), extensions, retain, qos)

        if "stats" in self.extensions:
            self.publish_statastics(retain, qos)

        if "firmware" in self.extensions:
            self.publish_firmware(retain, qos)

    def publish_firmware(self, retain = True, qos = 1):
        mac,ip = self.mqtt_client.get_mac_ip_address()

        self.publish("/".join((self.topic, "$localip")),ip, retain, qos)
        self.publish("/".join((self.topic, "$mac")),mac, retain, qos)
        self.publish("/".join((self.topic, "$fw/name")),self.homie_settings ['fw_name'], retain, qos)
        self.publish("/".join((self.topic, "$fw/version")),self.homie_settings ['fw_version'], retain, qos)
        self.publish("/".join((self.topic, "$implementation")),self.homie_settings ['implementation'], retain, qos)

    def publish_statastics(self, retain = True, qos = 1):
        self.publish("/".join((self.topic, "$stats/interval")),self.homie_settings ['update_interval'], retain, qos)
        self.publish("/".join((self.topic, "$stats/uptime")),time.time()-self.start_time, retain, qos)
        self.publish("/".join((self.topic, "$stats/lastupdate")),datetime.now().strftime("%d/%m/%Y %H:%M:%S"), retain, qos)

    def publish_uptime(self, retain=True, qos=1):
        self.publish("/".join((self.topic, "$stats/uptime")),time.time()-self.start_time, retain, qos)
        self.publish("/".join((self.topic, "$stats/lastupdate")),datetime.now().strftime("%d/%m/%Y %H:%M:%S"), retain, qos)

    def publish_homeassistant(self,hass_config,hass_payload):
        self.publish(hass_config,hass_payload, True, 1)
        
    def add_subscription(self,topic,handler,qos=0): #subscription list to the required MQTT topics, used by properties to catch set topics
        self.mqtt_subscription_handlers [topic] = handler
        self.mqtt_client.subscribe (topic,qos)
        logger.debug ('MQTT subscribed to {}'.format(topic))

    def remove_subscription(self,topic):
        self.mqtt_client.unsubscribe (topic)
        del self.mqtt_subscription_handlers [topic]
        logger.debug ('MQTT unsubscribed to {}'.format(topic))

    def subscribe_topics(self):
        logger.debug("Device subscribing to topics")
        self.add_subscription(
            "/".join((self.topic, "$broadcast/#")), self.broadcast_handler
        )  # get the broadcast events

        for _, node in self.nodes.items():
            for topic, handler in node.get_subscriptions().items():
                self.add_subscription(topic, handler)

    def add_node(self, node):
        self.nodes[node.id] = node

        if self.nodes_published:  # update, publish property changes
            self.publish_nodes(self.retained, self.qos)

    def remove_node(self, node_id):  # not tested, needs work removing topics
        del self.nodes[node_id]

        if self.nodes_published:  # update, publish property changes
            self.publish_nodes(retain=False)

    def get_node(self, node_id):
        if node_id in self.nodes:
            return self.nodes[node_id]
        else:
            return None

    def publish_nodes(self, retain=True, qos=1):
        nodes = ",".join(self.nodes.keys())
        self.publish("/".join((self.topic, "$nodes")), nodes, retain, qos)

        self.nodes_published = True

        for _, node in self.nodes.items():
            node.publish_attributes(retain, qos)

    def broadcast_handler(self, topic, payload):  # TBD
        logger.debug(
            "Device MQTT Homie Broadcast:  Topic {}, Payload {}".format(topic, payload)
        )

    def publish(self, topic, payload, retain, qos):
        logger.debug(
            "Device MQTT publish topic: {}, retain {}, qos {}, payload: {}".format(
                topic, retain, qos, payload
            )
        )
        self.mqtt_client.publish(topic, payload, retain=retain, qos=qos)

    def _homie_validate_settings(self, settings):
        if settings is not None:
            for setting, value in HOMIE_SETTINGS.items():
                logger.debug("Homie settings {} {}".format(setting, value))
                if not setting in settings:
                    settings[setting] = HOMIE_SETTINGS[setting]
        else:
            settings = HOMIE_SETTINGS

        return settings

    def mqtt_on_connection(self, connected):
        logger.info("Device MQTT Connected state is {}".format(connected))
        # print("Device MQTT Connected state is {}".format(connected))

        if connected:
            if self._mqtt_connected is False:
                self._mqtt_connected = True
                self.publish_attributes()
                self.publish_nodes()
                self.subscribe_topics()
                self.publish_homeassistant()
                if (
                    self.mqtt_client.using_shared_mqtt_client is False
                    or self.instance_number == 1
                ):  # only set last will if NOT using shared client or if using shared client and this is the first device instance
                    self.mqtt_client.set_will(
                        "/".join((self.topic, "$state")), "lost", retain=True, qos=1
                    )
                    logger.warning(
                        "Device setting last will {}".format(
                            "/".join((self.topic, "$state"))
                        )
                    )
        else:
            self._mqtt_connected = False

    def mqtt_on_message(self, topic, payload, retain, qos):
        if topic in self.mqtt_subscription_handlers:
            logger.debug(
                "Device MQTT Message: Topic {}, Payload {} Retain {}  QOS {}".format(
                    topic, payload, retain, qos
                )
            )  # for logging only, topic and handler for subsriptions above

            if (
                retain is False
            ):  # OH2.5 MQTT sends set messages with retain true - need to check to avoid this problem
                self.mqtt_subscription_handlers[topic](topic, payload)
            else:
                logger.warn(
                    "Device MQTT Message received with RETAIN TRUE: Topic {}, Payload {} Retain {}  QOS {}".format(
                        topic, payload, retain, qos
                    )
                )  # for logging only, topic and handler for subsriptions above


def close_devices(*arg):
    logger.info ('Closing Devices')
    global devices 
    for device in devices:
        device.close()
    logger.info ('Closed Devices')

    close_mqtt_clients()

#atexit.register(close_devices)

#signal.signal(signal.SIGTERM, close_devices)
#signal.signal(signal.SIGINT, close_devices)




from .node_base import Node_Base

from homie.node.property.property_integer import Property_Integer


class Node_Integer(Node_Base):
    def __init__(
        self,
        device,
        id="integer",
        name="State",
        type_="integer",
        retain=True,
        qos=1,
        set_value=None,
    ):
        super().__init__(device, id, name, type_, retain, qos)

        assert set_value

        self.add_property(
            Property_Integer(self, "integer", "Integer", set_value=set_value)
        )

    def update_value(self, value):
        self.get_property("integer").value = value




import time
from threading import Event, Thread
import traceback

import logging

logger = logging.getLogger(__name__)


class Repeating_Timer(object):
    DEFAULT_INTERVAL_REDUCTION = 0.9

    """Repeat `function` every `interval` seconds."""

    def __init__(self, interval):
        self.interval = int(interval * self.DEFAULT_INTERVAL_REDUCTION)

        self.start = time.time()
        self.event = Event()

        self.thread = Thread(target=self._target)
        self.thread.setDaemon(True)
        self.thread.start()

        self.callbacks = []

    def _target(self):
        while not self.event.wait(self._time):
            for callback in self.callbacks:
                try:
                    callback()
                except Exception as e:
                    logger.error("Error in timer callback: {}  {}".format(e,traceback.format_exc()))

    @property
    def _time(self):
        return self.interval - ((time.time() - self.start) % self.interval)

    def add_callback(self, callback):
        self.callbacks.append(callback)

    def stop(self):
        self.event.set()
        self.thread.join()




from .property_boolean import Property_Boolean


tags = ["Switch"]


class Property_Switch(Property_Boolean):
    def __init__(
        self,
        node,
        id="switch",
        name="Switch",
        settable=True,
        retained=True,
        qos=1,
        unit=None,
        data_type=None,
        data_format=None,
        value=None,
        set_value=None,
        tags=tags,
        meta={},
    ):
        super().__init__(
            node,
            id,
            name,
            settable,
            retained,
            qos,
            unit,
            data_type,
            data_format,
            value,
            set_value,
            tags,
            meta,
        )

    def validate_value(self, value):
        return value in ["ON", "OFF"]

    def get_value_from_payload(self, payload):
        if payload == "true":
            return "ON"
        elif payload == "false":
            return "OFF"
        else:
            return None

    def get_payload_from_value(self, value):
        if value == "ON":
            return "true"
        else:
            return "false"




from .property_base import Property_Base

import datetime


class Property_DateTime(Property_Base):
    def __init__(
        self,
        node,
        id,
        name,
        settable=False,
        retained=True,
        qos=1,
        unit=None,
        data_type="datetime",
        data_format=None,
        value=None,
        set_value=None,
        tags=[],
        meta={},
    ):

        super().__init__(
            node,
            id,
            name,
            settable,
            retained,
            qos,
            unit,
            "datetime",
            data_format,
            value,
            set_value,
            tags,
            meta,
        )

    def validate_value(self, value):
        try:
            date_time_obj = datetime.datetime.strptime(
                date_time_str, "%Y-%m-%d %H:%M:%S.%f"
            )
            return True
        except:
            return False

    def get_value_from_payload(self, payload):
        try:
            return datetime.datetime.strptime(payload, "%Y-%m-%d %H:%M:%S.%f")
        except:
            return None

